@@ -10,7 +10,7 @@ Copyright 1999, 2000, 2001 Regents of the University of California.
 Released for unlimited redistribution; see file Legal.htm
 Documentation is in the Numeric manual; see numpy.sourceforge.net
 """
-import Numeric, PropertiedClasses
+import Numeric
 import string, types, sys
 from Precision import *
 from Numeric import e, pi, NewAxis
@@ -481,6 +481,8 @@ subtract = masked_binary_operation(Numeric.subtract)
 subtract.reduce = None
 multiply = masked_binary_operation(Numeric.multiply, 1, 1)
 divide = domained_binary_operation(Numeric.divide, domain_safe_divide(), 0, 1)
+true_divide = domained_binary_operation(Numeric.true_divide, domain_safe_divide(), 0, 1)
+floor_divide = domained_binary_operation(Numeric.floor_divide, domain_safe_divide(), 0, 1)
 remainder = domained_binary_operation(Numeric.remainder, domain_safe_divide(), 0, 1)
 fmod = domained_binary_operation(Numeric.fmod, domain_safe_divide(), 0, 1)
 hypot = masked_binary_operation(Numeric.hypot)
@@ -508,7 +510,7 @@ rank = Numeric.rank
 shape = Numeric.shape
 size = Numeric.size
 
-class MaskedArray (PropertiedClasses.PropertiedClass):
+class MaskedArray (object):
     """Arrays with possibly masked values. 
        Masked values of 1 exclude element from the computation.
        Construction:
@@ -553,13 +555,15 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
         need_data_copied = copy
         if isinstance(data, MaskedArray): 
             c = data.raw_data()
+            ctc = c.typecode()
             if tc is None:
-                tc = c.typecode()
-            elif tc != c.typecode():
+                tc = ctc
+            elif tc != ctc:
                 need_data_copied = 1
+            css = c.spacesaver()
             if ss is None:
-                ss = c.spacesaver()
-            elif ss != c.spacesaver():
+                ss = css
+            elif ss != css:
                 need_data_copied = 1
             else:
                 ss = 0
@@ -570,23 +574,28 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
 
         elif isinstance(data, Numeric.ArrayType): 
             c = data
+            ctc = c.typecode()
             if tc is None:
-                tc = c.typecode()
-            elif tc != c.typecode():
+                tc = ctc
+            elif tc != ctc:
                 need_data_copied = 1
+            css = c.spacesaver()
             if ss is None:
-                ss = c.spacesaver()
-            elif ss != c.spacesaver():
+                ss = css
+            elif ss != css:
                 need_data_copied = 1
         else:
-            c = data
-            need_data_copied = 1
+            need_data_copied = 0 #because I'll do it now
             if ss is None: 
                 ss = 0
+            c = Numeric.array(data, tc, savespace=ss)
 
         if need_data_copied:
-            self._data = Numeric.array(c, typecode=tc, copy=1)
-            self.savespace(ss)
+            if tc == ctc:
+                self._data = Numeric.array(c, copy=1, savespace = ss)
+            else:
+                self._data = c.astype(tc)
+                self._data.savespace(ss)
         else:
             self._data = c
 
@@ -595,21 +604,24 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
             self._shared_mask = 0
         else:
             self._mask = make_mask (mask, flag=ss)
-            self._shared_mask = (self._mask is mask)
-            nm = size(self._mask)
-            nd = size(self._data)
-            if nm != nd:
-                if nm == 1:
-                    self._mask = Numeric.resize(self._mask, self._data.shape)
-                    self._shared_mask = 0
-                elif nd == 1:
-                    self._data = Numeric.resize(self._data, self._mask.shape)
-                    self._data.shape = self._mask.shape
-                else:
-                    raise MAError, "Mask and data not compatible."
-            elif nm == 1 and shape(self._mask) != shape(self._data):
-                self.unshare_mask()
-                self._mask.shape = self._data.shape
+            if self._mask is None:
+                self._shared_mask = 0
+            else:
+                self._shared_mask = (self._mask is mask)
+                nm = size(self._mask)
+                nd = size(self._data)
+                if nm != nd:
+                    if nm == 1:
+                        self._mask = Numeric.resize(self._mask, self._data.shape)
+                        self._shared_mask = 0
+                    elif nd == 1:
+                        self._data = Numeric.resize(self._data, self._mask.shape)
+                        self._data.shape = self._mask.shape
+                    else:
+                        raise MAError, "Mask and data not compatible."
+                elif nm == 1 and shape(self._mask) != shape(self._data):
+                    self.unshare_mask()
+                    self._mask.shape = self._data.shape
 
         self.set_fill_value(fill_value)
 
@@ -630,11 +642,11 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
         else:
             return self._data
     
-    def _get_shape(self, name):
+    def _get_shape(self):
         "Return the current shape."
         return self._data.shape
 
-    def _set_shape (self, name, newshape):
+    def _set_shape (self, newshape):
         "Set the array's shape."
         if not self._data.iscontiguous():
             self._data = Numeric.array(self._data, self._data.typecode(), 
@@ -646,7 +658,7 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
                 self._mask = Numeric.array(self._mask, MaskType, 1, 1)
             self._mask.shape = newshape
             
-    def _get_flat(self, name):
+    def _get_flat(self):
         """Calculate the flat value. 
         """
         if self._mask is None:
@@ -657,12 +669,12 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
                                 mask=self._mask.flat,
                                 fill_value = self.fill_value())
 
-    def _set_flat (self, name, value):
+    def _set_flat (self, value):
         "x.flat = value"
         y = self.flat
         y[:] = value
 
-    def _get_real(self, name):
+    def _get_real(self):
         "Get the real part of a complex array."
         if self._mask is None:
             return masked_array(self._data.real, mask=None,
@@ -671,12 +683,12 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
             return masked_array(self._data.real, mask=self.mask().flat,
                             fill_value = self.fill_value())
     
-    def _set_real (self, name, value):
+    def _set_real (self, value):
         "x.real = value"
         y = self.real
         y[...] = value
 
-    def _get_imaginary(self, name):
+    def _get_imaginary(self):
         "Get the imaginary part of a complex array."
         if self._mask is None:
             return masked_array(self._data.imaginary, mask=None,
@@ -685,7 +697,7 @@ class MaskedArray (PropertiedClasses.PropertiedClass):
             return masked_array(self._data.imaginary, mask=self.mask().flat,
                             fill_value = self.fill_value())
 
-    def _set_imaginary (self, name, value):
+    def _set_imaginary (self, value):
         "x.imaginary = value"
         y = self.imaginary
         y[...] = value
@@ -749,13 +761,17 @@ array(data = %(data)s,
 
     def __float__(self):
         "Convert self to float."
-        x = filled(self, 0)
-        return masked_array(self.astype(Float), self.mask())[...]
+        self.unmask()
+        if self._mask is not None:
+            raise MAError, 'Cannot convert masked element to a Python float.'
+        return float(self.raw_data()[...])
 
     def __int__(self):
         "Convert self to int."
-        x = filled(self, 0)
-        return masked_array(self.astype(Int), self.mask())[...]
+        self.unmask()
+        if self._mask is not None:
+            raise MAError, 'Cannot convert masked element to a Python int.'
+        return int(self.raw_data()[...])
 
 # Note copy semantics here differ from Numeric        
     def __getitem__(self, i): 
@@ -917,6 +933,22 @@ array(data = %(data)s,
         "Return divide(other, self)"
         return divide(other, self)
 
+    def __truediv__(self, other): 
+        "Return divide(self, other)"
+        return true_divide(self, other)
+
+    def __rtruediv__(self, other): 
+        "Return divide(other, self)"
+        return true_divide(other, self)
+
+    def __floordiv__(self, other): 
+        "Return divide(self, other)"
+        return floor_divide(self, other)
+
+    def __rfloordiv__(self, other): 
+        "Return divide(other, self)"
+        return floor_divide(other, self)
+
     def __pow__(self,other, third=None): 
         "Return power(self, other, third)"
         return power(self, other, third)
@@ -968,6 +1000,49 @@ array(data = %(data)s,
             self._shared_mask = 1
         return self
 
+    def __imul__(self, other): 
+        "Add other to self in place."
+        t = self._data.typecode()
+        f = filled(other,0)
+        t1 = f.typecode()
+        if t == t1:
+            pass
+        elif t in typecodes['Integer']:
+            if t1 in typecodes['Integer']:
+                f = f.astype(t)
+            else:
+                raise TypeError, 'Incorrect type for in-place operation.'
+        elif t in typecodes['Float']:
+            if t1 in typecodes['Integer']:
+                f = f.astype(t)
+            elif t1 in typecodes['Float']:
+                f = f.astype(t)
+            else:
+                raise TypeError, 'Incorrect type for in-place operation.'
+        elif t in typecodes['Complex']:
+            if t1 in typecodes['Integer']:
+                f = f.astype(t)
+            elif t1 in typecodes['Float']:
+                f = f.astype(t)
+            elif t1 in typecodes['Complex']:
+                f = f.astype(t)
+            else:
+                raise TypeError, 'Incorrect type for in-place operation.'
+        else:
+            raise TypeError, 'Incorrect type for in-place operation.'
+
+        if self._mask is None:
+            self._data *= f
+            m = getmask(other)
+            self._mask = m
+            self._shared_mask = m is not None
+        else:
+            result = multiply(self, masked_array(f, mask=getmask(other)))
+            self._data = result.raw_data()
+            self._mask = result.raw_mask()
+            self._shared_mask = 1
+        return self
+
     def __isub__(self, other): 
         "Subtract other from self in place."
         t = self._data.typecode()
@@ -1011,48 +1086,7 @@ array(data = %(data)s,
             self._shared_mask = 1
         return self
 
-    def __imul__(self, other): 
-        "Multiply self by other in place."
-        t = self._data.typecode()
-        f = filled(other,0)
-        t1 = f.typecode()
-        if t == t1:
-            pass
-        elif t in typecodes['Integer']:
-            if t1 in typecodes['Integer']:
-                f = f.astype(t)
-            else:
-                raise TypeError, 'Incorrect type for in-place operation.'
-        elif t in typecodes['Float']:
-            if t1 in typecodes['Integer']:
-                f = f.astype(t)
-            elif t1 in typecodes['Float']:
-                f = f.astype(t)
-            else:
-                raise TypeError, 'Incorrect type for in-place operation.'
-        elif t in typecodes['Complex']:
-            if t1 in typecodes['Integer']:
-                f = f.astype(t)
-            elif t1 in typecodes['Float']:
-                f = f.astype(t)
-            elif t1 in typecodes['Complex']:
-                f = f.astype(t)
-            else:
-                raise TypeError, 'Incorrect type for in-place operation.'
-        else:
-            raise TypeError, 'Incorrect type for in-place operation.'
 
-        if self._mask is None:
-            self._data *= f
-            m = getmask(other)
-            self._mask = m
-            self._shared_mask = m is not None
-        else:
-            result = multiply(self, masked_array(f, mask=getmask(other)))
-            self._data = result.raw_data()
-            self._mask = result.raw_mask()
-            self._shared_mask = 1
-        return self
 
     def __idiv__(self, other): 
         "Divide self by other in place."
@@ -1084,20 +1118,13 @@ array(data = %(data)s,
                 raise TypeError, 'Incorrect type for in-place operation.'
         else:
             raise TypeError, 'Incorrect type for in-place operation.'
-
-        if self._mask is None:
-            dm = domain_safe_divide()(self._data, filled(other,0))
-            self._data /= f
+        mo = getmask(other)
+        result = divide(self, masked_array(f, mask=mo))
+        self._data = result.raw_data()
+        dm = result.raw_mask()
+        if dm is not self._mask:
             self._mask = dm
-            self._shared_mask = dm is not None
-        else:
-            mo = getmask(other)
-            result = divide(self, masked_array(f, mask=mo))
-            self._data = result.raw_data()
-            dm = result.raw_mask()
-            if dm is not self._mask:
-                self._mask = dm
-                self._shared_mask = 1
+            self._shared_mask = 1
         return self
 
     def __eq__(self,other): 
@@ -1325,32 +1352,19 @@ array(data = %(data)s,
             self._mask = make_mask (self._mask, copy=1, flag=0)
             self._shared_mask = 0
 
+    shape = property(_get_shape, _set_shape,
+           doc = 'tuple giving the shape of the array')
 
-PropertiedClasses.set_property (MaskedArray, 'shape', 
-                                MaskedArray._get_shape,
-                                MaskedArray._set_shape,
-                                nodelete=1
-             )
-PropertiedClasses.set_property (MaskedArray, 'flat', 
-                              MaskedArray._get_flat, 
-                              MaskedArray._set_flat, 
-                              nodelete=1
-                             )
-PropertiedClasses.set_property (MaskedArray, 'real', 
-                              MaskedArray._get_real, 
-                              MaskedArray._set_real,
-                              nodelete = 1
-                             ) 
-PropertiedClasses.set_property (MaskedArray, 'imaginary', 
-                              MaskedArray._get_imaginary, 
-                              MaskedArray._set_imaginary,
-                              nodelete = 1
-                             ) 
-PropertiedClasses.set_property (MaskedArray, 'imag', 
-                              MaskedArray._get_imaginary, 
-                              MaskedArray._set_imaginary,
-                              nodelete = 1
-                             ) 
+    flat = property(_get_flat, _set_flat,
+           doc = 'Access array in flat form.')
+
+    real = property(_get_real, _set_real,
+           doc = 'Access the real part of the array')
+
+    imaginary = property(_get_imaginary, _set_imaginary,
+           doc = 'Access the imaginary part of the array')
+
+    imag = imaginary
                                                         
 array = MaskedArray
 
@@ -1360,11 +1374,8 @@ class MaskedScalar (MaskedArray):
         self._data.shape = ()
         self._mask.shape = ()
 
-PropertiedClasses.set_property(MaskedScalar, 'shape', nowrite=1)
-PropertiedClasses.set_property(MaskedScalar, 'imaginary', nowrite=1)
-PropertiedClasses.set_property(MaskedScalar, 'imag', nowrite=1)
-PropertiedClasses.set_property(MaskedScalar, 'real', nowrite=1)
-PropertiedClasses.set_property(MaskedScalar, 'flat', nowrite=1)
+    shape = property(MaskedArray._get_shape)
+
 masked = MaskedScalar()
 
 def isMaskedArray (x):
@@ -1579,56 +1590,95 @@ def average (a, axis=0, weights=None, returned = 0):
 
        masked values in the weights will be set to 0.0
     """
-    a = array(a)
-    ma = getmaskarray(a)
+    a = masked_array(a)
+    mask = a.mask()
+    ash = a.shape
+    if ash == ():
+        ash = (1,)
     if axis is None:
-        if weights is None:
-            w = masked_array(Numeric.choose(ma, (1.0,0.0)), ma)
-        else:
-            w = array(filled(weights, 0.0), Float, mask = ma, copy = 0)
-        n = add.reduce(a.flat*w.flat)
-        d = add.reduce(w.flat) 
-    else:
-        ash = a.shape
-        if ash == ():
-            a.shape = (1,)
-        if weights is None:
-            w = masked_array(Numeric.choose(ma, (1.0,0.0)), ma)
-            n = add.reduce(a*w, axis)
-            d = add.reduce(w, axis)
-        else:
-            wv = filled(weights, 0.0) 
-            wsh = wv.shape
-            if wsh == ():
-                wsh = (1,)
-            if wsh == ash:
-                w = array(wv, Float, mask=ma, copy=0)
-                n = add.reduce(a*w, axis)
-                d = add.reduce(w, axis) 
-            elif wsh == (ash[axis],):
-                ni = ash[axis]
-                r = [NewAxis]*len(ash)
-                r[axis] = slice(None,None,1)
-                w = eval ("wv["+ repr(tuple(r)) + "] * masked_array(ones(ash, Float), ma)")
-                n = add.reduce(a*w, axis)
-                d = add.reduce(w, axis)
+        if mask is None:
+            if weights is None:
+                n = add.reduce(a.raw_data().flat)
+                d = reduce(lambda x, y: x * y, ash, 1.0)
             else:
-                raise ValueError, 'average: weights wrong shape.'
+                w = filled(weights, 0.0).flat
+                n = Numeric.add.reduce(a.raw_data().flat * w)
+                d = Numeric.add.reduce(w)
+                del w
+        else:
+            if weights is None:
+                n = add.reduce(a.flat)
+                w = Numeric.choose(mask, (1.0,0.0)).flat
+                d = Numeric.add.reduce(w)
+                del w
+            else:
+                w = array(filled(weights, 0.0), Numeric.Float, mask=mask).flat
+                n = add.reduce(a.flat * w)
+                d = add.reduce(w)
+                del w
+    else:
+        if mask is None:
+            if weights is None:
+                d = ash[axis] * 1.0
+                n = Numeric.add.reduce(a.raw_data(), axis)
+            else:
+                w = filled(weights, 0.0) 
+                wsh = w.shape
+                if wsh == ():
+                    wsh = (1,)
+                if wsh == ash:
+                    w = Numeric.array(w, Float, copy=0)
+                    n = add.reduce(a*w, axis)
+                    d = add.reduce(w, axis) 
+                    del w
+                elif wsh == (ash[axis],):
+                    ni = ash[axis]
+                    r = [NewAxis]*len(ash)
+                    r[axis] = slice(None,None,1)
+                    w = eval ("w["+ repr(tuple(r)) + "] * ones(ash, Float)")
+                    n = add.reduce(a*w, axis)
+                    d = add.reduce(w, axis)
+                    del w, r
+                else:
+                    raise ValueError, 'average: weights wrong shape.'
+        else:
+            if weights is None:
+                n = add.reduce(a, axis)
+                w = Numeric.choose(mask, (1.0, 0.0))
+                d = Numeric.add.reduce(w, axis)
+                del w
+            else:
+                w = filled(weights, 0.0) 
+                wsh = w.shape
+                if wsh == ():
+                    wsh = (1,)
+                if wsh == ash:
+                    w = array(w, Float, mask=mask, copy=0)
+                    n = add.reduce(a*w, axis)
+                    d = add.reduce(w, axis) 
+                elif wsh == (ash[axis],):
+                    ni = ash[axis]
+                    r = [NewAxis]*len(ash)
+                    r[axis] = slice(None,None,1)
+                    w = eval ("w["+ repr(tuple(r)) + "] * masked_array(ones(ash, Float), mask)")
+                    n = add.reduce(a*w, axis)
+                    d = add.reduce(w, axis)
+                else:
+                    raise ValueError, 'average: weights wrong shape.'
+                del w
+    # print n, d, repr(mask), repr(weights)
+    result = divide (n, d)
+    del n
 
-    result = n/d
-    # This routine is a very hard routine to get right -- PFD
-    # Keep these debug prints as comments for future suffering.
-    #print "a", a
-    #print "ma", ma
-    #print "w", w
-    #print "n", n
-    #print "d", d            
-    #print "r", result
     if isinstance(result, MaskedArray):
-        result.unmask()   
-    if returned:
-        if isinstance(result, MaskedArray):
+        result.unmask() 
+        if returned:  
+            if not isinstance(d, MaskedArray):
+                d = masked_array(d)
+            if not d.shape == result.shape:
+                d = ones(result.shape, Float) * d
             d.unmask()
+    if returned:
         return result, d
     else:
         return result
@@ -1705,13 +1755,19 @@ def masked_less_equal(x, value, copy=1):
 
 def masked_not_equal(x, value, copy=1):
     "masked_not_equal(x, value) = x masked where x != value"
-    return masked_where(not_equal(x, value), x, copy)
+    d = filled(x,0)
+    c = Numeric.not_equal(d, value)
+    m = mask_or(c, getmask(x))
+    return array(d, mask=m, copy=copy)
 
 def masked_equal(x, value, copy=1):
     """masked_equal(x, value) = x masked where x == value
        For floating point consider masked_values(x, value) instead.
     """
-    return masked_where(equal(x, value), x, copy)
+    d = filled(x,0)
+    c = Numeric.equal(d, value)
+    m = mask_or(c, getmask(x))
+    return array(d, mask=m, copy=copy)
 
 def masked_inside(x, v1, v2, copy=1):
     """x with mask of all values of x that are inside [v1,v2]
@@ -1721,10 +1777,10 @@ def masked_inside(x, v1, v2, copy=1):
         t = v2
         v2 = v1
         v1 = t
-    d = filled(x)
-    c = logical_and(less_equal(d, v2), greater_equal(d, v1))
+    d=filled(x, 0)
+    c = Numeric.logical_and(Numeric.less_equal(d, v2), Numeric.greater_equal(d, v1))
     m = mask_or(c, getmask(x))
-    return array(d, mask = c, copy=copy)
+    return array(d, mask = m, copy=copy)
 
 def masked_outside(x, v1, v2, copy=1):
     """x with mask of all values of x that are outside [v1,v2]
@@ -1734,10 +1790,10 @@ def masked_outside(x, v1, v2, copy=1):
         t = v2
         v2 = v1
         v1 = t
-    d = filled(x)
-    c = logical_or(less(d, v1), greater(d, v2))
+    d = filled(x,0)
+    c = Numeric.logical_or(Numeric.less(d, v1), Numeric.greater(d, v2))
     m = mask_or(c, getmask(x))
-    return array(d, mask = c, copy=copy)
+    return array(d, mask = m, copy=copy)
 
 def reshape (a, newshape):
     "Copy of a with a new shape."
