@@ -776,3 +776,15 @@ def average (a, axis=0, weights=None, returned = 0):
         return n/d, d
     else:
         return n/d
+
+def vdot(a, b):
+    """Returns the dot product of 2 vectors (or anything that can be made into
+       a vector). NB: this is not the same as `dot`, as it takes the conjugate
+       of its first argument if complex and always returns a scalar."""
+    return multiarray.matrixproduct(conjugate(ravel(a)),
+                                    ravel(b))
+
+# try to import blas optimized dot, innerproduct and vdot, if available
+try:
+    from dotblas import dot, innerproduct, vdot
+except ImportError: pass
